// My_cluster class declaration and definition
//
// by Gianluca Aglieri Rinella
//
// last modified on 26/8/04

// My_cluster class DECLARATION and IMPLEMENTATION
// software by Gianluca Aglieri Rinella
// Doctoral Student at CERN
// European Organization for Nuclear Research
//
// with permission to use it freely but reproducing this comment
// including when reused, follow GPL and I will be happy
//
// last modified on 01/09/05
//
// My_cluster is a C++ class to store and work with vectors of My_whit objects
// modelling weighted hit clusters
// 
// class My_hit is a simple but useful hit description class
// it stores a hit by its row and col coordinates
// wrapped together
//
// class My_whit adds to My_hit the capacity to store a weight for a hit,
// i.e. the number of times the pixel was hit.
//
// to use these classes include this file from your source
// this software is using only very standard C++ classes and include files
//

// class My_cluster
//
//  manages a vector of adiacent My_whit objects
// It is a simple container but takes care of cluster statistics
// calculations, like total weight, center of mass, average row and col...
//
//
//  Interface:
//
//  inline unsigned int Size();  returns number of My_whits in My_cluster
//  int Push(My_whit& new_hit);  add a new My_whit object. IMPORTANT:
//                               the new My_whit is added ONLY if
//                               it is the first hit, or if adiacent
//                               to one of the hits already in the cluster
//  
//  int Pop();                   kicks out last hit of the cluster
//  My_whit& operator[] (int i); gives back the i-th hit in the cluster
// 
//  int Push_force(My_whit& new_hit); pushes in the hit with no check
//                                    on adiacence, see the comment to
//                                    the Push member function 
//  getters of statistics
//  inline double GetSumW();   sum of weights
//  inline double GetAvgX();   average column coordinate
//  inline double GetAvgY();   average row coordinate
//  inline double GetSigmaX(); root mean square deviation of column coordinate
//  inline double GetSigmaY(); root mean square deviation of row coordinates


#ifndef MY_CLUSTER
#define MY_CLUSTER

#include <vector>
#include <iostream>
#include <math.h>
#include "./My_whit.C" //needs My_whit.C and My_hit.C

using namespace std;


class My_cluster {

  vector<My_whit> hits_vector; // private data structure 
                               // is a vector of My_whit
                               // i.e. weighted hits
  bool updated;    // to keep trace of the need to re-calculate
                   // statistics. Statistics are calculated only when 
                   // needed
 
  // statistics storage space:
  // sum_w    sum weights
  // sum_wx   sum weight*col
  // sum_wxx  sum weights*col*col
  // ...
  double sum_w, sum_wx, sum_wy, sum_wxx, sum_wyy;
  double avg_x, avg_y;
  double sigma_x, sigma_y;
  

  // a private member method  to refresh statistics 
  void refresh() {
    double w, x, y;

    sum_w = 0.0;
      sum_wx = 0.0;
      sum_wy = 0.0;
      sum_wxx = 0.0;
      sum_wyy = 0.0;

    for(unsigned int i=0;i<hits_vector.size(); i++) {
      w = hits_vector[i].GetWeight();  // upcasting here!
      x = hits_vector[i].GetCol();     // yep also
      y = hits_vector[i].GetRow();     // sure also here
      sum_w += w;
      sum_wx += w*x;
      sum_wy += w*y;
      sum_wxx += w*x*x;
      sum_wyy += w*y*y;
    }

    if(sum_w==0.0) {
      avg_x =0.0; avg_y=0.0; sigma_x=0.0; sigma_y=0.0;
      updated = true;
    }
    else {
      avg_x = sum_wx / sum_w + 0.5; // the 0.5 is due to the necessity of shifting to the center of the pixel position
      avg_y = sum_wy / sum_w + 0.5;
      sigma_x = sqrt (sum_wxx/sum_w + avg_x - 1/6 - avg_x*avg_x); // again: work out the calculation for the sigma considering the proper 
      // definition; should come out that formula
      sigma_y = sqrt (sum_wyy/sum_w + avg_y - 1/6 - avg_y*avg_y);
      updated = true;  // now statistics are refreshed
    }
  }


  // interface: Size(), Push(), Pop(), are folding vector template equivalents
  // but Push() checks if the new hit can be in the cluster
  // operator [] is also defined for the class (which is a kind of sophisticated
  // container in the end)

public :
  My_cluster() { // default constructor
    updated=true; 
    sum_w =0.0; sum_wx=0.0; sum_wy=0.0; sum_wxx=0.0; sum_wyy=0.0;
    avg_x=0.0; avg_y=0.0; sigma_x=0.0; sigma_y=0.0;
}


  // gives back number of My_whits in the cluster
  inline unsigned int Size() { return hits_vector.size(); }


  // Push to add a new hit at end of vector; 
  // need to check if can be in the cluster, i.e. if new hit is
  // adiacent to any of the hits already in the cluster;
  // also checks if not already in the cluster
  // gives back cluster size.
  // add the hit without check to an empty cluster

  int Push(My_whit& new_hit) {   // by reference for efficiency ?! maybe?!
    //   cout << "Called Push" << endl;
    bool found = false, same = false;
    for(unsigned int i=0; i<hits_vector.size() ;i++) { // scan the vector
      found = hits_vector[i].IsAdiacentTo(new_hit) ; // found an adiacent?
      same = hits_vector[i].IsTheSame(new_hit); // found the same??
    }

    // add the hit only if found an adiacent and not already in the cluster, or if empty cluster
    if((found==true && same == false) || hits_vector.size()==0 ) {
      hits_vector.push_back(new_hit);
      updated = false;  // added a new hit, so need statistics are not
      // anymore up to date 
    }
    return hits_vector.size(); 
  }

int Push_force(My_whit& new_hit) {   // by reference for efficiency ?
				     // guess so
  // this does not check for consistency like Push function does.
    //   cout << "Called Push" << endl;
    bool same = false;
    for(unsigned int i=0; i<hits_vector.size() ;i++) {
      // found = hits_vector[i].IsAdiacentTo(new_hit) ;
      same = hits_vector[i].IsTheSame(new_hit);
    }
    // add the hit only if not already in the cluster, or if empty cluster
    if((same == false) || hits_vector.size()==0 ) {
      hits_vector.push_back(new_hit);
      updated = false;  // added a new hit, so need to update statistics
    }
    return hits_vector.size(); 
  }

  // this is to pop out last hit in the vector; will pop lat hit from
  // vector and unset the updated flag  
  int Pop() {
    if(hits_vector.size()>0) {
      hits_vector.pop_back();
      updated = false;
    }
    return hits_vector.size();
  }

  // overloading operator [] to access hits in the cluster
  My_whit& operator[] (unsigned int i) {
    if(i>=0 && i<hits_vector.size()) {
      return hits_vector[i];
    }
    else {
      cout << i << " index out of cluster size range" << endl;
      return hits_vector[hits_vector.size()-1];
    }
  }

// getters of statistics in the interface. CHeck first if statistics
//  are up to date. If not recalculate by the refresh method.
  inline double GetSumW() {
   if (updated == false) { refresh(); }
    return sum_w;
  } 

  inline double GetAvgX() {
    if (updated == false) { refresh(); }
    return avg_x;
  }

  inline double GetAvgY() {
    if (updated == false) { refresh(); }
    return avg_y;
  }

  inline double GetSigmaX() {
    if (updated == false) { refresh(); }
    return sigma_x;
  }

inline double GetSigmaY() {
    if (updated == false) { refresh(); }
    return sigma_y;
  }


};  // end of My_cluster class


// VERY useful tool
// a clustering function
// wrapped in clustering_tools namespace
// call it with
// cluster_list clustering_tools::Clustering(hit_list);
// 
// takes as input a vector of My_whit and gives back
// a vector of My_cluster objects (all the clusters found)
//
// there is anoverloaded version working also with vector of 
// My_hit as argument


namespace clustering_tools {

vector<My_cluster > Clustering (vector<My_whit > hits_list) {
  
  // let's define the cluster list
  vector<My_cluster > clusters_list;

  // let us make sure there are no pathologies like two hits
  // with same coordinates. You never know..
  // I passed by value, so this change will not affect calling
  // function, hits_list is in fact local
  My_whit_tools::Eliminate_twins(hits_list);

  // now starts the clustering
  // first phase: scan all hits and put them into clusters
  //

  bool found = false; // to see if we found a cluster for the hit

  // scan over all hits
  for (unsigned int i=0; i<hits_list.size();i++) {

    found = false; // starting with a new hit not found a cluster for
                   // it yet

    //scan over all clusters founded till now
    for(unsigned int j=0; j<clusters_list.size() && found==false; j++) {

      // store initial j-th cluster size
      int old_cluster_size = clusters_list[j].Size();  // Size() method of My_cluster class
      // try to Push the My_whit into the j-th cluster
      // this already checks for adjacent hits in the cluster!!
      int cluster_size = clusters_list[j].Push(hits_list[i]);
      // see if we the cluster was ok for the i-th hit
      if(cluster_size == old_cluster_size+1) {
	//no need to add explicitly the hit, already done if it was ok
	found = true; // flag that the i-th hit was put in a connected set
      }
    }

    // we now scanned all clusters, if not found a cluster for the hit, then
    // add a cluster at the end of the list of clusters, with the new hit as
    // single hit

    if (found==false) {
      My_cluster new_cluster;
      new_cluster.Push(hits_list[i]);
      clusters_list.push_back(new_cluster);
    }


  } // let us go to next hit (i+1)


  // end of first phase
  // second phase now: merging of clusters
  //
  // there's now the need to merge the found clusters 
  // if they have adiacent hits
   // in fact it may happen we have not grouped properly due to the unknown
  // ordering of the hits in the list

    //cycle from end to beginning of list of clusters
  for(int i = clusters_list.size()-1; i>0; i--) { // scan backwards

    found = false;
    //cycle over all clusters preceeding the i-th
    for(int j=0; j<i && found == false; j++) {
      
      
	//check if j-th and i-th have some adiacent pixels
      for(unsigned int ni=0; ni<clusters_list[i].Size() && found == false; ni++) {
	for (unsigned int nj=0; nj<clusters_list[j].Size() && found == false; nj++) {
	  found = clusters_list[i][ni].IsAdiacentTo(clusters_list[j][nj]);
	}
      }
      // if found at least one couple of adiacent hits, 
      // queue i-th cluster to j-th, need
      // to force the push of the hit without checking adiacency
      // I declare that if I am not wrong,
      // the two clusters have not twin hits, so there is no risk of
      // redundant insertions of hits
      // But we are obliged to use the forced push, cause the hits in
      // j-th clusters are not necessarily DIRECTLY adiacent to those
      // of the i-th
 
      if (found == true ) {
	for(unsigned int pos = 0; pos < clusters_list[i].Size(); pos++) {
	  clusters_list[j].Push_force(clusters_list[i][pos]);
	}
	// now kick away the i-th cluster from list
	// i-th is not necesarily last, so got to shift all the
	// following leftwards before kicking
	// out the last one 
	for(unsigned int cc=i+1;cc<clusters_list.size();cc++){
	  clusters_list[cc-1]=clusters_list[cc];
	}
	clusters_list.pop_back();
      }

    }
  }

  return clusters_list; // return list of clusters

} //end of Clustering function



// Let us overload the previous function to work also with a vector of My_hit

  vector<My_cluster > Clustering (vector<My_hit > &hits_list) {
    vector<My_whit > weighted_list;
    // copy the list of My_hit into a list of My_whit with unitarian weights
    for(unsigned int i= 0; i< hits_list.size(); i++) {
      weighted_list.push_back(My_whit(hits_list[i].GetRow(),hits_list[i].GetCol(),1));
    }

    return Clustering(weighted_list);
  }

}

#endif
